#include "bd.h"

#include <algorithm>

bool BD::Better::operator()(BD* l, BD* r) {
  return l->score() < r->score();
}

BD::BD(FILE* fp)
  : num_lambda_(0),
    lambda_(0),
    total_lambda_(0),
    score_(0),
    done_(0),
    x_(0),
    y_(0),
    w_(0),
    h_(0),
    turn_(0),
    water_(0),
    flooding_(0),
    waterproof_(10),
    life_(1) {
  char* buf = NULL;
  size_t buf_len;
  ssize_t len;
  while ((len = getline(&buf, &buf_len, fp)) >= 0) {
    assert(len);
    if (buf[len-1] == '\n') {
      buf[len-1] = '\0';
    }
    string line = buf;
    if (line.size() == 0)
      break;
    mine_.push_back(line);
    w_ = max<int>(w_, line.size());
  }
  free(buf);
  reverse(mine_.begin(), mine_.end());

  if (len >= 0) {
    char buf[256];
    int value;
    for (;;) {
      int r = fscanf(fp, "%s %d\n", buf, &value);
      if (r == -1)
        break;
      if (r != 2) {
        fprintf(stderr, "invalid command\n");
        abort();
      }
      if (!strcmp(buf, "Water")) {
        water_ = value;
      } else if (!strcmp(buf, "Flooding")) {
        flooding_ = value;
      } else if (!strcmp(buf, "Waterproof")) {
        waterproof_ = value;
      } else {
        fprintf(stderr, "invalid command: %s\n", buf);
        abort();
      }
    }

    life_ = waterproof_;
  }

  assert(!mine_.empty());
  h_ = (int)mine_.size();

  for (int i = 0; i < h_; i++) {
    size_t found = mine_[i].find('R');
    if (found != string::npos) {
      x_ = found;
      y_ = i;
    }

    found = mine_[i].find_first_of("LO");
    if (found != string::npos) {
      gx_ = found;
      gy_ = i;
    }

    for (size_t j = 0; j < mine_[i].size(); j++) {
      if (mine_[i][j] == '*') {
        rocks_.push_back(make_pair(j, i));
      }
    }
  }

  calcNumLambda();

  fprintf(stderr,
          "x=%d y=%d w=%d h=%d lambda=%d "
          "water=%d flooding=%d waterproof=%d\n",
          x_, y_, w_, h_, num_lambda_,
          water_, flooding_, waterproof_);
}

void BD::show(FILE* o) const {
#ifdef NOLOG
  if (o == stderr)
    return;
#endif
  fprintf(o, "     ");
  for (int x = 0; x < w_; x++) {
    fprintf(o, "%d", x % 10);
  }
  fprintf(o, "\n");
  int level = getWaterLevel() - 1;
  for (int y = (int)mine_.size() - 1; y >= 0; y--) {
    fprintf(o, "%4d %s", y, mine_[y].c_str());
    if (y == level) {
      fprintf(o, " ~\n");
    } else {
      fprintf(o, "\n");
    }
  }
}

void BD::addScore(int score) {
  score_ += score;
}

void BD::move(int dx, int dy, int* rx) {
  int nx = x_ + dx;
  int ny = y_ + dy;
  if (ny < 0 || ny >= (int)mine_.size() ||
      nx < 0 || nx >= (int)mine_[ny].size())
    return;
  switch (mine_[ny][nx]) {
  case '\\':
    lambda_++;
    total_lambda_++;
    addScore(25);
  case ' ':
  case '.':
    break;
  case 'O':
    done_ = GOAL;
    addScore(total_lambda_ * 50);
    break;
  case '*':
    if (dy == 0 && mine_[ny][nx+dx] == ' ') {
      mine_[ny][nx+dx] = '*';
      *rx = nx+dx;
      break;
    } else {
      return;
    }
  default:
    return;
  }

  mine_[y_][x_] = ' ';
  y_ = ny;
  x_ = nx;
  mine_[y_][x_] = 'R';
}

void BD::tick(char c) {
  turn_++;
  seq_.push_back(c);

  int rx = -1;
  switch (c) {
  case 'L':
    move(-1, 0, &rx);
    break;
  case 'R':
    move(1, 0, &rx);
    break;
  case 'U':
    move(0, 1, &rx);
    break;
  case 'D':
    move(0, -1, &rx);
    break;
  case 'W':
    break;
  case 'A':
    done_ = ABORT;
    addScore(total_lambda_ * 25);
    return;
  default:
    fprintf(stderr, "unknown command %c (%d)\n", c, c);
    abort();
  }

  score_--;

  vector<pair<pair<int, int>, pair<int, int> > > dropped;
  for (pair<int, int>& rock : rocks_) {
    int x = rock.first;
    int y = rock.second;
    if (x == x_ && y == y_) {
      assert(rx >= 0);
      x = rock.first = rx;
    }

    if (mine_[y-1][x] == ' ') {
      y--;
    } else if (mine_[y-1][x] == '*') {
      if (mine_[y][x+1] == ' ' && mine_[y-1][x+1] == ' ') {
        y--;
        x++;
      } else if (mine_[y][x-1] == ' ' && mine_[y-1][x-1] == ' ') {
        y--;
        x--;
      }
    } else if (mine_[y-1][x] == '\\') {
      if (mine_[y][x+1] == ' ' && mine_[y-1][x+1] == ' ') {
        y--;
        x++;
      }
    }

    if (rock.first != x || rock.second != y) {
      dropped.push_back(make_pair(rock, make_pair(x, y)));
      rock.first = x;
      rock.second = y;
    }
  }

  for (const pair<pair<int, int>, pair<int, int> >& rock : dropped) {
    {
      const pair<int, int>& before = rock.first;
      int x = before.first;
      int y = before.second;
      mine_[y][x] = ' ';
    }
    {
      const pair<int, int>& after = rock.second;
      int x = after.first;
      int y = after.second;
      mine_[y][x] = 'v';
    }
  }

#if 0
  for (int y = 0; y < h_; y++) {
    for (int x = 0; x < (int)mine_[y].size(); x++) {
      mine_work_[y][x] = mine_[y][x];
      if (mine_[y][x] == '*') {
        if (mine_[y-1][x] == ' ') {
          mine_work_[y-1][x] = '*';
          mine_work_[y][x] = ' ';
        } else if (mine_[y-1][x] == '*') {
          if (mine_[y][x+1] == ' ' && mine_[y-1][x+1] == ' ') {
            mine_work_[y-1][x+1] = '*';
            mine_work_[y][x] = ' ';
          } else if (mine_[y][x-1] == ' ' && mine_[y-1][x-1] == ' ') {
            mine_work_[y-1][x-1] = '*';
            mine_work_[y][x] = ' ';
          }
        } else if (mine_[y-1][x] == '\\') {
          if (mine_[y][x+1] == ' ' && mine_[y-1][x+1] == ' ') {
            mine_work_[y-1][x+1] = '*';
            mine_work_[y][x] = ' ';
          }
        }
      } else if (mine_[y][x] == 'L' && lambda_ == num_lambda_) {
        mine_work_[y][x] = 'O';
      }
    }
  }
#endif

  if (lambda_ == num_lambda_) {
    mine_[gy_][gx_] = 'O';
  }

  if (y_ < getWaterLevel()) {
    life_--;
  } else {
    life_ = waterproof_;
  }

  // win
  if (done_)
    return;

  // lose
  if ((y_ < (int)mine_.size() - 1 && mine_[y_+1][x_] == 'v') ||
      life_ < 0) {
    done_ = DEAD;
    return;
  }

  for (const pair<pair<int, int>, pair<int, int> >& rock : dropped) {
    const pair<int, int>& after = rock.second;
    int x = after.first;
    int y = after.second;
    mine_[y][x] = '*';
  }
}

void BD::calcNumLambda() {
  num_lambda_ = 0;
  for (int i = 0; i < h_; i++) {
    for (char c : mine_[i]) {
      if (c == '\\')
        num_lambda_++;
    }
  }
}

bool BD::getGoal(int* x, int* y) const {
#if 0
  for (int i = 0; i < h_; i++) {
    size_t found = mine_[i].find('O');
    if (found != string::npos) {
      *x = found;
      *y = i;
      return true;
    }
  }
  return false;
#endif
  *x = gx_;
  *y = gy_;
  return get(gx_, gy_) == 'O';
}

bool BD::makeGoal(int* x, int* y) {
  for (int i = 0; i < h_; i++) {
    size_t found = mine_[i].find_first_of("OL");
    if (found != string::npos) {
      *x = found;
      *y = i;
      mine_[*y][*x] = 'O';
      return true;
    }
  }
  return false;
}

int BD::getDistance(int gx, int gy) const {
  return abs(x_ - gx) + abs(y_ - gy);
}
