#include "bd.h"

#include <limits.h>

#include <queue>
#include <set>

int main(int argc, char* argv[]) {
  if (argc < 2) {
    fprintf(stderr, "%s <map> [ctrl]\n", argv[0]);
    abort();
  }

  FILE* map_fp = fopen(argv[1], "rb");
  assert(map_fp);

  BD* bd = new BD(map_fp);
  fclose(map_fp);

  priority_queue<BD*, vector<BD*>, BD::Better> q;
  q.push(bd);

  const string cmds = "LRUDWA";

  int best_score = INT_MIN;
  BD* best_bd = NULL;
  set<vector<string> > seen;
  while (!q.empty()) {
    BD* bd = q.top();
    q.pop();

    //bd->show();

    if (!seen.insert(bd->mine()).second) {
      delete bd;
      continue;
    }

    for (char c : cmds) {
      BD* nbd = new BD(*bd);
      nbd->tick(c);
      if (nbd->done()) {
        if (best_score < nbd->score()) {
          best_score = nbd->score();
          best_bd = nbd;
          fprintf(stderr, "score=%d seq=%s\n",
                  best_bd->score(), best_bd->seq().c_str());
        } else {
          delete nbd;
        }
      } else {
        q.push(nbd);
      }
    }
    delete bd;
  }

  fprintf(stderr, "score=%d\n", best_bd->score());
  printf("%s\n", best_bd->seq().c_str());
}
